//------------------------------------------------------------------------------
// File: CompressorPlugin.cpp
//------------------------------------------------------------------------------

/*
 based on AudioCompress v1.5 (2002/07/31)
 (c)2002 trikuare studios (http://trikuare.cx)

 http://trikuare.cx/code/AudioCompress.html
*/

/*
* Strongly modified for ReClock by OGO :)
*
*/

#include <streams.h>
#include <initguid.h>

#include "CompressorPlugin.h"

#undef COMPRESSOR_STATS

#define COMPRESSOR_BUCKETS 20		/* How long of a history to store */
#define COMPRESSOR_TARGET 25000		/* Target level */
#define COMPRESSOR_GAINSHIFT 10		/* How fine-grained the gain is */

#define COMPRESSOR_GAINMAX_LIGHT		5	/* The maximum amount to amplify by */
#define COMPRESSOR_GAINSMOOTH_LIGHT		8	/* How much inertia ramping has*/
#define COMPRESSOR_BUCKETS_TIME_LIGHT	3	/* Length of history in seconds */

#define COMPRESSOR_GAINMAX_NORMAL		5	/* The maximum amount to amplify by */
#define COMPRESSOR_GAINSMOOTH_NORMAL	2	/* How much inertia ramping has*/
#define COMPRESSOR_BUCKETS_TIME_NORMAL	1	/* Length of history in seconds */

#define COMPRESSOR_GAINMAX_STRONG		6	/* The maximum amount to amplify by */
#define COMPRESSOR_GAINSMOOTH_STRONG	1	/* How much inertia ramping has*/
#define COMPRESSOR_BUCKETS_TIME_STRONG	1	/* Length of history in seconds */

#define SAFE_DELETE(p)       { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }

Compressor::Compressor()
{
	m_peaks = 0;
	SetParameters(44100, DRC_NORMAL);
}

void Compressor::SetParameters(int samplesPerSec, int effect)
{
	int l_bucketsTime = 1;

	m_anticlip = false;
	m_target = COMPRESSOR_TARGET;

	switch(effect)
	{
		case DRC_LIGHT:
			m_gainmax = COMPRESSOR_GAINMAX_LIGHT;
			m_gainsmooth = COMPRESSOR_GAINSMOOTH_LIGHT;
			l_bucketsTime = COMPRESSOR_BUCKETS_TIME_LIGHT;
			break;
			
		case DRC_NORMAL:
			m_gainmax = COMPRESSOR_GAINMAX_NORMAL;
			m_gainsmooth = COMPRESSOR_GAINSMOOTH_NORMAL;
			l_bucketsTime = COMPRESSOR_BUCKETS_TIME_NORMAL;
			break;
			
		case DRC_STRONG:
		default:
			m_gainmax = COMPRESSOR_GAINMAX_STRONG;
			m_gainsmooth = COMPRESSOR_GAINSMOOTH_STRONG;
			l_bucketsTime = COMPRESSOR_BUCKETS_TIME_STRONG;
			break;
	}
	
	m_gainCurrent = m_gainTarget = (1 << COMPRESSOR_GAINSHIFT);
	
	m_clip = 0;
	
	/* Allocate peaks */
	SAFE_DELETE_ARRAY(m_peaks);
	m_peaks = new long[COMPRESSOR_BUCKETS];
	for (int i = 0; i < COMPRESSOR_BUCKETS; i++)
		m_peaks[i] = 0;
	m_samplesInCurrentBucket = 0;
	m_currentBucketPeak = 1;
	m_samplesPerBucket = (samplesPerSec * l_bucketsTime) / COMPRESSOR_BUCKETS;
	m_pn = -1;
}

Compressor::~Compressor()
{
	SAFE_DELETE_ARRAY(m_peaks);
}

double Compressor::GetCurrentGain() 
{ 
	return (double)m_gainCurrent / (double)(1 << COMPRESSOR_GAINSHIFT); 
}

void Compressor::ProcessFlow32(long *input, long *output, int sampleCount)
{
	/* Determine peak's value and position */
	long l_peak = 1<<16;
	int l_pos = 0;
	
	long* l_ap = input;
	for (int i = 0; i < sampleCount; i++)
	{
		long l_val = *l_ap++;
		
		if (l_val > l_peak)
		{
			l_peak = l_val;
			l_pos = i;
		} 
		else if (-l_val > l_peak)
		{
			l_peak = -l_val;
			l_pos = i;
		}

		if (l_val > m_currentBucketPeak)
			m_currentBucketPeak = l_val;
		else if (-l_val > m_currentBucketPeak)
			m_currentBucketPeak = -l_val;
		if (++m_samplesInCurrentBucket >= m_samplesPerBucket)
		{
			m_pn = (m_pn + 1) % COMPRESSOR_BUCKETS;
			m_peaks[m_pn] = m_currentBucketPeak;
			m_currentBucketPeak = 1;
			m_samplesInCurrentBucket = 0;
		}
	}
	
	for (int i = 0; i < COMPRESSOR_BUCKETS; i++)
	{
		if (m_peaks[i] > l_peak)
		{
			l_peak = m_peaks[i];
			l_pos = 0;
		}
	}
	
	/* Determine target gain */
	int l_gn = ((1 << COMPRESSOR_GAINSHIFT) * m_target) / (l_peak>>16);
	if (l_gn < (1 << COMPRESSOR_GAINSHIFT))
		l_gn = 1 << COMPRESSOR_GAINSHIFT;
	
	m_gainTarget = (m_gainTarget * ((1 << m_gainsmooth) - 1) + l_gn) >> m_gainsmooth;
	
	/* Give it an extra insignifigant nudge to counteract possible
	** rounding error
	*/
	
	if (l_gn < m_gainTarget)
		m_gainTarget--;
	else if (l_gn > m_gainTarget)
		m_gainTarget++;
	
	if (m_gainTarget > m_gainmax << COMPRESSOR_GAINSHIFT)
		m_gainTarget = m_gainmax << COMPRESSOR_GAINSHIFT;
	
	/* See if a peak is going to clip */
	l_gn = ((1 << COMPRESSOR_GAINSHIFT) * 32768) / (l_peak>>16);
	
	if (l_gn < m_gainTarget)
	{
		m_gainTarget = l_gn;
		
		if (m_anticlip)
			l_pos = 0;
		
	} else
	{
		/* We're ramping up, so draw it out over the whole frame */
		l_pos = (m_samplesPerBucket * COMPRESSOR_BUCKETS);
	}
	
	/* Determine gain rate necessary to make target */
	if (l_pos == 0)
		l_pos = 1;
	
	int l_gr = ((m_gainTarget - m_gainCurrent) << 16) / l_pos;
	
	/* Do the shiznit */
	int l_gf = m_gainCurrent << 16;
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( " gain = %2.2f %+.2e ", m_gainCurrent*1.0/(1 << COMPRESSOR_GAINSHIFT), (m_gainTarget - m_gainCurrent)*1.0/(1 << COMPRESSOR_GAINSHIFT));
#endif
	
	l_ap = input;
	for (int i = 0; i < sampleCount; i++)
	{
		/* Interpolate the gain */
		m_gainCurrent = l_gf >> 16;
		if (i < l_pos)
			l_gf += l_gr;
		else if (i == l_pos)
			l_gf = m_gainTarget << 16;
		
		/* Amplify */
		_int64 l_sample = ((_int64)(*l_ap++) * (_int64)m_gainCurrent) >> COMPRESSOR_GAINSHIFT;
		if (l_sample < -(_int64)2147483648)
		{
			m_clip++;
			l_sample = -(_int64)2147483648;
		}
		else if (l_sample > (_int64)2147483647)
		{
			m_clip++;
			l_sample = (_int64)2147483647;
		}

		*output++ = (long)l_sample;
	}
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( "clip %d b%-3d\n", m_clip, m_pn);
#endif
}

#define GETSAMPLE24(adr) ((((unsigned char *)adr)[0] << 8) | (((unsigned char *)adr)[1] << 16) | (((unsigned char *)adr)[2] << 24))
#define SETSAMPLE24(adr, val) { *((unsigned char*)adr+0) = *((unsigned char*)(&val)+1) ; *((unsigned char*)adr+1) = *((unsigned char*)(&val)+2); *((unsigned char*)adr+2) = *((unsigned char*)(&val)+3); }

void Compressor::ProcessFlow24(void *input, void *output, int sampleCount)
{
	/* Determine peak's value and position */
	long l_peak = 1<<16;
	int l_pos = 0;
	
	void* l_ap = input;
	for (int i = 0; i < sampleCount; i++)
	{
		long l_val = GETSAMPLE24(l_ap);
		l_ap = (char*)l_ap + 3;
		
		if (l_val > l_peak)
		{
			l_peak = l_val;
			l_pos = i;
		} 
		else if (-l_val > l_peak)
		{
			l_peak = -l_val;
			l_pos = i;
		}

		if (l_val > m_currentBucketPeak)
			m_currentBucketPeak = l_val;
		else if (-l_val > m_currentBucketPeak)
			m_currentBucketPeak = -l_val;
		if (++m_samplesInCurrentBucket >= m_samplesPerBucket)
		{
			m_pn = (m_pn + 1) % COMPRESSOR_BUCKETS;
			m_peaks[m_pn] = m_currentBucketPeak;
			m_currentBucketPeak = 1;
			m_samplesInCurrentBucket = 0;
		}
	}
	
	for (int i = 0; i < COMPRESSOR_BUCKETS; i++)
	{
		if (m_peaks[i] > l_peak)
		{
			l_peak = m_peaks[i];
			l_pos = 0;
		}
	}
	
	/* Determine target gain */
	int l_gn = ((1 << COMPRESSOR_GAINSHIFT) * m_target) / (l_peak>>16);
	if (l_gn < (1 << COMPRESSOR_GAINSHIFT))
		l_gn = 1 << COMPRESSOR_GAINSHIFT;
	
	m_gainTarget = (m_gainTarget * ((1 << m_gainsmooth) - 1) + l_gn) >> m_gainsmooth;
	
	/* Give it an extra insignifigant nudge to counteract possible
	** rounding error
	*/
	
	if (l_gn < m_gainTarget)
		m_gainTarget--;
	else if (l_gn > m_gainTarget)
		m_gainTarget++;
	
	if (m_gainTarget > m_gainmax << COMPRESSOR_GAINSHIFT)
		m_gainTarget = m_gainmax << COMPRESSOR_GAINSHIFT;
	
	/* See if a peak is going to clip */
	l_gn = ((1 << COMPRESSOR_GAINSHIFT) * 32768) / (l_peak>>16);
	
	if (l_gn < m_gainTarget)
	{
		m_gainTarget = l_gn;
		
		if (m_anticlip)
			l_pos = 0;
		
	} else
	{
		/* We're ramping up, so draw it out over the whole frame */
		l_pos = (m_samplesPerBucket * COMPRESSOR_BUCKETS);
	}
	
	/* Determine gain rate necessary to make target */
	if (l_pos == 0)
		l_pos = 1;
	
	int l_gr = ((m_gainTarget - m_gainCurrent) << 16) / l_pos;
	
	/* Do the shiznit */
	int l_gf = m_gainCurrent << 16;
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( " gain = %2.2f %+.2e ", m_gainCurrent*1.0/(1 << COMPRESSOR_GAINSHIFT), (m_gainTarget - m_gainCurrent)*1.0/(1 << COMPRESSOR_GAINSHIFT));
#endif
	
	l_ap = input;
	for (int i = 0; i < sampleCount; i++)
	{
		/* Interpolate the gain */
		m_gainCurrent = l_gf >> 16;
		if (i < l_pos)
			l_gf += l_gr;
		else if (i == l_pos)
			l_gf = m_gainTarget << 16;
	
		/* Amplify */
		long l_val = GETSAMPLE24(l_ap);
		l_ap = (char*)l_ap + 3;
		_int64 l_sample = ((_int64)l_val * (_int64)m_gainCurrent) >> COMPRESSOR_GAINSHIFT;
		
		if (l_sample < -(_int64)2147483648)
		{
			m_clip++;
			l_sample = -(_int64)2147483648;
		}
		else if (l_sample > (_int64)2147483647)
		{
			m_clip++;
			l_sample = (_int64)2147483647;
		}

		SETSAMPLE24(output, l_sample);
		output = (char*)output + 3;
	}
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( "clip %d b%-3d\n", m_clip, m_pn);
#endif
}

void Compressor::ProcessFlow16(short *input, short *output, int sampleCount)
{
	/* Determine peak's value and position */
	int l_peak = 1;
	int l_pos = 0;
	
	short* l_ap = input;
	for (int i = 0; i < sampleCount; i++)
	{
		int l_val = *l_ap++;
		
		if (l_val > l_peak)
		{
			l_peak = l_val;
			l_pos = i;
		} 
		else if (-l_val > l_peak)
		{
			l_peak = -l_val;
			l_pos = i;
		}

		if (l_val > m_currentBucketPeak)
			m_currentBucketPeak = l_val;
		else if (-l_val > m_currentBucketPeak)
			m_currentBucketPeak = -l_val;
		if (++m_samplesInCurrentBucket >= m_samplesPerBucket)
		{
			m_pn = (m_pn + 1) % COMPRESSOR_BUCKETS;
			m_peaks[m_pn] = m_currentBucketPeak;
			m_currentBucketPeak = 1;
			m_samplesInCurrentBucket = 0;
		}
	}
	
	for (int i = 0; i < COMPRESSOR_BUCKETS; i++)
	{
		if (m_peaks[i] > l_peak)
		{
			l_peak = m_peaks[i];
			l_pos = 0;
		}
	}
	
	/* Determine target gain */
	int l_gn = ((1 << COMPRESSOR_GAINSHIFT) * m_target) / l_peak;
	if (l_gn < (1 << COMPRESSOR_GAINSHIFT))
		l_gn = 1 << COMPRESSOR_GAINSHIFT;
	
	m_gainTarget = (m_gainTarget * ((1 << m_gainsmooth) - 1) + l_gn) >> m_gainsmooth;
	
	/* Give it an extra insignifigant nudge to counteract possible
	** rounding error
	*/
	
	if (l_gn < m_gainTarget)
		m_gainTarget--;
	else if (l_gn > m_gainTarget)
		m_gainTarget++;
	
	if (m_gainTarget > m_gainmax << COMPRESSOR_GAINSHIFT)
		m_gainTarget = m_gainmax << COMPRESSOR_GAINSHIFT;
	
	/* See if a peak is going to clip */
	l_gn = ((1 << COMPRESSOR_GAINSHIFT) * 32768) / l_peak;
	
	if (l_gn < m_gainTarget)
	{
		m_gainTarget = l_gn;
		
		if (m_anticlip)
			l_pos = 0;
		
	} else
	{
		/* We're ramping up, so draw it out over the whole frame */
		l_pos = (m_samplesPerBucket * COMPRESSOR_BUCKETS);
	}
	
	/* Determine gain rate necessary to make target */
	if (l_pos == 0)
		l_pos = 1;
	
	int l_gr = ((m_gainTarget - m_gainCurrent) << 16) / l_pos;
	
	/* Do the shiznit */
	int l_gf = m_gainCurrent << 16;
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( " gain = %2.2f %+.2e ", m_gainCurrent*1.0/(1 << COMPRESSOR_GAINSHIFT), (m_gainTarget - m_gainCurrent)*1.0/(1 << COMPRESSOR_GAINSHIFT));
#endif
	
	l_ap = input;
	for (int i = 0; i < sampleCount; i++)
	{
		/* Interpolate the gain */
		m_gainCurrent = l_gf >> 16;
		if (i < l_pos)
			l_gf += l_gr;
		else if (i == l_pos)
			l_gf = m_gainTarget << 16;
		
		/* Amplify */
		long l_sample = ((long)(*l_ap++) * (long)m_gainCurrent) >> COMPRESSOR_GAINSHIFT;
		if (l_sample < -32768)
		{
			m_clip++;
			l_sample = -32768;
		}
		else if (l_sample > 32767)
		{
			m_clip++;
			l_sample = 32767;
		}

		*output++ = (short)l_sample;
	}
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( "clip %d b%-3d\n", m_clip, m_pn);
#endif
}

void Compressor::ProcessFlow8(unsigned char *input, unsigned char *output, int sampleCount)
{
	/* Determine peak's value and position */
	long l_peak = 1;
	int l_pos = 0;
	
	unsigned char* l_ap = input;
	for (int i = 0; i < sampleCount; i++)
	{
		int l_val = (int)(*l_ap++) - 128;
		
		if (l_val > l_peak)
		{
			l_peak = l_val;
			l_pos = i;
		} 
		else if (-l_val > l_peak)
		{
			l_peak = -l_val;
			l_pos = i;
		}

		if (l_val > m_currentBucketPeak)
			m_currentBucketPeak = l_val;
		else if (-l_val > m_currentBucketPeak)
			m_currentBucketPeak = -l_val;
		if (++m_samplesInCurrentBucket >= m_samplesPerBucket)
		{
			m_pn = (m_pn + 1) % COMPRESSOR_BUCKETS;
			m_peaks[m_pn] = m_currentBucketPeak;
			m_currentBucketPeak = 1;
			m_samplesInCurrentBucket = 0;
		}
	}
	
	for (int i = 0; i < COMPRESSOR_BUCKETS; i++)
	{
		if (m_peaks[i] > l_peak)
		{
			l_peak = m_peaks[i];
			l_pos = 0;
		}
	}
	
	/* Determine target gain */
	int l_gn = ((1 << COMPRESSOR_GAINSHIFT) * m_target) / (l_peak<<8);
	if (l_gn < (1 << COMPRESSOR_GAINSHIFT))
		l_gn = 1 << COMPRESSOR_GAINSHIFT;
	
	m_gainTarget = (m_gainTarget * ((1 << m_gainsmooth) - 1) + l_gn) >> m_gainsmooth;
	
	/* Give it an extra insignifigant nudge to counteract possible
	** rounding error
	*/
	
	if (l_gn < m_gainTarget)
		m_gainTarget--;
	else if (l_gn > m_gainTarget)
		m_gainTarget++;
	
	if (m_gainTarget > m_gainmax << COMPRESSOR_GAINSHIFT)
		m_gainTarget = m_gainmax << COMPRESSOR_GAINSHIFT;
	
	/* See if a peak is going to clip */
	l_gn = ((1 << COMPRESSOR_GAINSHIFT) * 32768) / (l_peak<<8);
	
	if (l_gn < m_gainTarget)
	{
		m_gainTarget = l_gn;
		
		if (m_anticlip)
			l_pos = 0;
		
	} else
	{
		/* We're ramping up, so draw it out over the whole frame */
		l_pos = (m_samplesPerBucket * COMPRESSOR_BUCKETS);
	}
	
	/* Determine gain rate necessary to make target */
	if (l_pos == 0)
		l_pos = 1;
	
	int l_gr = ((m_gainTarget - m_gainCurrent) << 16) / l_pos;
	
	/* Do the shiznit */
	int l_gf = m_gainCurrent << 16;
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( " gain = %2.2f %+.2e ", m_gainCurrent*1.0/(1 << COMPRESSOR_GAINSHIFT), (m_gainTarget - m_gainCurrent)*1.0/(1 << COMPRESSOR_GAINSHIFT));
#endif
	
	l_ap = (unsigned char*)input;
	for (int i = 0; i < sampleCount; i++)
	{
		/* Interpolate the gain */
		m_gainCurrent = l_gf >> 16;
		if (i < l_pos)
			l_gf += l_gr;
		else if (i == l_pos)
			l_gf = m_gainTarget << 16;
		
		/* Amplify */
		int l_sample = (((int)(*l_ap++) - 128) * (int)m_gainCurrent) >> COMPRESSOR_GAINSHIFT;
		if (l_sample < -128)
		{
			m_clip++;
			l_sample = -128;
		}
		else if (l_sample > 127)
		{
			m_clip++;
			l_sample = 127;
		}

		*output++ = (unsigned char)(l_sample+128);
	}
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( "clip %d b%-3d\n", m_clip, m_pn);
#endif
}

